"""
This examples shows how Asynker can be used in a more complex setting.

This is a graphical tool which can fetch a website and show the size of
embedded resources (such as scripts, stylesheets and images) -- it's almost
bordering on being useful for something.

PyQt5 and lxml are required.
"""

import sys

from PyQt5.QtCore import QObject, pyqtSlot, QEventLoop, QUrl, QThread, pyqtSignal, Q_ARG, Qt
from PyQt5.QtWidgets import QMainWindow, QVBoxLayout, QLineEdit, QPushButton, \
    QApplication, QWidget, QTableWidget, QTableWidgetItem, QHeaderView
from PyQt5.QtNetwork import QNetworkAccessManager, QNetworkRequest

import asynker
import lxml.html


class NetworkRequestFailed(RuntimeError):
    error: int
    error_string: str

    def __init__(self, error, error_string):
        self.error = error
        self.error_string = error_string
        super().__init__('%s (%d)' % (error_string, error))


class Network(QObject):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.network_access = QNetworkAccessManager(parent=self)
        self.in_flight = {}  # QNetworkReply -> asynker.Future

    @pyqtSlot()
    def run(self):
        self.sched = asynker.Scheduler()
        self.run_event_loop()

        print('Shutting event loop down.')
        for reply in list(self.in_flight):
            reply.abort()  # This will trigger the respective finished() signals, which will cancel related futures
        self.sched.cancel_all_tasks()
        self.sched.run_until_all_tasks_finished()
        print('Event loop shut down.')

    def run_event_loop(self):
        """
        Runs the event loop. Will only complete if the thread receives an interruption request
        (QThread.requestInterruption()).

        Qt manages this thread's event loop, meanwhile asynker will dispatch completed futures to their
        respective coroutines.
        """
        thread = self.thread()
        while not thread.isInterruptionRequested():
            self.sched.tick()
            thread.eventDispatcher().processEvents(QEventLoop.AllEvents|QEventLoop.WaitForMoreEvents)

    def get(self, url: QUrl) -> asynker.Future:
        future = asynker.Future()
        request = QNetworkRequest(url)
        reply = self.network_access.get(request)
        reply.finished.connect(self.network_request_finished)
        self.in_flight[reply] = future
        return future

    @pyqtSlot()
    def network_request_finished(self):
        reply = self.sender()
        future = self.in_flight.pop(reply)

        if reply.error():
            exc = NetworkRequestFailed(reply.error(), reply.errorString())
            future.set_exception(exc)
        else:
            future.set_result(reply)


class PageStats(QObject):
    done = pyqtSignal()
    result = pyqtSignal(dict)
    summary = pyqtSignal(dict)

    def __init__(self, network, parent=None):
        super().__init__(parent)
        self.network = network
        self.future = None

    @pyqtSlot(QUrl)
    def fetch_stats(self, url):
        if not self.future or self.future.done():
            self.future = self.network.sched.run(self._fetch_stats(url))
            self.future.add_done_callback(self._done)
        else:
            print('Still working.')

    async def _fetch_stats(self, url):
        def emit_result(result):
            summary['requests'] += 1
            summary['size'] += result['size']
            self.result.emit(result)

        summary = dict(requests=0, size=0)
        reply = await self.network.get(url)
        body = reply.readAll().data()
        emit_result(dict(type='page', path='.', url=url, size=len(body)))

        # For simplicity, assume UTF-8.
        tree = lxml.html.document_fromstring(body.decode())
        ext_resources = self._extract_resources(tree, url)

        async for resource in asynker.as_completed(*[self._get_resource_size(r) for r in ext_resources], scheduler=self.network.sched):
            emit_result(resource)

        self.summary.emit(summary)

    async def _get_resource_size(self, resource):
        reply = await self.network.get(resource['url'])
        resource['size'] = len(reply.readAll().data())
        return resource

    def _extract_resources(self, etree, base_url):
        def get_elements_with_tag(tag):
            return etree.xpath('.//' + tag)

        def add_resource(type, path):
            resources.append(dict(type=type, path=path, url=base_url.resolved(QUrl(path))))

        resources = []

        for element in get_elements_with_tag('link'):
            if element.attrib.get('rel', None) == 'stylesheet' and 'href' in element.attrib:
                add_resource('stylesheet', element.attrib['href'])
        for element in get_elements_with_tag('script'):
            if 'src' in element.attrib:
                add_resource('script', element.attrib['src'])
        for element in get_elements_with_tag('img'):
            if element.attrib.get('src'):
                add_resource('image', element.attrib['src'])

        return resources

    def _done(self, future):
        if future.exception():
            print('Exception in fetch_stats:', future.exception())
        self.done.emit()


class MainWindow(QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setMinimumWidth(700)
        self.setMinimumHeight(500)

        self.workerthread = QThread(self)
        self.network = Network()
        self.network.moveToThread(self.workerthread)
        self.pagestats = PageStats(self.network)
        self.pagestats.moveToThread(self.workerthread)
        self.workerthread.started.connect(self.network.run)
        self.workerthread.start()

        self.setCentralWidget(QWidget())
        self.centralWidget().setLayout(QVBoxLayout())
        layout = self.centralWidget().layout()

        self.urlinput = QLineEdit()
        self.urlinput.setPlaceholderText('Enter URL of web page here')
        self.dobutton = QPushButton('Fetch stats')
        self.dobutton.clicked.connect(self.do)
        self.results = QTableWidget(0, 3)
        self.results.setHorizontalHeaderLabels(['Type', 'Path', 'Size'])
        self.results.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self.results.setEditTriggers(QTableWidget.NoEditTriggers)

        self.pagestats.result.connect(self.show_result)
        self.pagestats.done.connect(self.done)
        self.pagestats.summary.connect(self.show_summary)

        layout.addWidget(self.urlinput)
        layout.addWidget(self.dobutton)
        layout.addWidget(self.results)

    def closeEvent(self, e):
        self.workerthread.requestInterruption()
        self.workerthread.eventDispatcher().wakeUp()
        self.workerthread.quit()
        self.workerthread.wait()
        super().closeEvent(e)

    @pyqtSlot()
    def do(self):
        url = QUrl.fromUserInput(self.urlinput.text())
        if url.isValid():
            self.metaObject().invokeMethod(self.pagestats, 'fetch_stats', Qt.QueuedConnection, Q_ARG(QUrl, url))
            self.dobutton.setEnabled(False)
            self.results.clearContents()
            self.results.setRowCount(0)
        else:
            self.statusBar().showMessage('Invalid url.')

    @pyqtSlot(dict)
    def show_result(self, result):
        row = self.results.rowCount()
        item1 = QTableWidgetItem(result['type'])
        item2 = QTableWidgetItem(result['path'])
        item3 = QTableWidgetItem(self.locale().formattedDataSize(result['size']))
        self.results.insertRow(row)
        self.results.setItem(row, 0, item1)
        self.results.setItem(row, 1, item2)
        self.results.setItem(row, 2, item3)

    @pyqtSlot(dict)
    def show_summary(self, summary):
        msg = '%s in %d requests' % (self.locale().formattedDataSize(summary['size']), summary['requests'])
        self.statusBar().showMessage(msg)

    @pyqtSlot()
    def done(self):
        self.dobutton.setEnabled(True)


if __name__ == '__main__':
    app = QApplication(sys.argv)
    mw = MainWindow()
    mw.show()
    app.exec()



import decimal  # accurate benchmarking requires accurate tools
from time import perf_counter

from asynker import Scheduler, suspend

N = 900000


async def loop():
    for n in range(N):
        await suspend()


if __name__ == '__main__':
    sched = Scheduler()
    t0 = perf_counter()
    sched.run_until_complete(loop())
    td = perf_counter() - t0

    decimal.getcontext().prec = 77  # industry-strength 256 full-bit-accuracy
    hz = decimal.Decimal(N) / decimal.Decimal(td)
    print('It\'s over %s yields per second!1' % hz)



import sys

from PyQt5.QtCore import QTimer, QEventLoop
from PyQt5.QtWidgets import QApplication, QDialog, QHBoxLayout, QPushButton, QLineEdit

from asynker import Scheduler, Future, suspend


app = QApplication(sys.argv)
window = QDialog()
window.setLayout(QHBoxLayout())
edit = QLineEdit()
window.layout().addWidget(edit)
button = QPushButton('Click me')
window.layout().addWidget(button)


def signal(sig):
    def emitted(*args):
        sig.disconnect(emitted)
        future.set_result(args)
    future = Future()
    sig.connect(emitted)
    return future


async def run_event_loop():
    while window.isVisible():
        await suspend()
        app.processEvents(QEventLoop.AllEvents | QEventLoop.WaitForMoreEvents)


async def dialogue():
    await signal(button.clicked)
    edit.setText('Type something here.')
    await signal(edit.editingFinished)
    button.setText('Very good.')

    timer = QTimer()
    timer.setSingleShot(True)
    timer.start(1500)
    await signal(timer.timeout)
    button.setText('Click me again to quit.')
    await signal(button.clicked)
    window.hide()


scheduler = Scheduler()
scheduler.run(dialogue())
window.show()
scheduler.run_until_complete(run_event_loop())



import pytest

from asynker import Scheduler, suspend, Future, CancelledError, gather, as_completed


def test_chain():
    flag = False

    async def other():
        nonlocal flag
        await suspend()
        flag = True
        return 42

    async def entry():
        return await other()

    sched = Scheduler()
    assert sched.run_until_complete(entry()) == 42
    assert flag


def test_chain_exc():
    async def other():
        raise KeyError('Foo')

    async def entry():
        return await other()

    sched = Scheduler()
    with pytest.raises(KeyError):
        sched.run_until_complete(entry())


def test_loop():
    f = Future()

    def get_response():
        return f

    async def entry():
        return await get_response()

    sched = Scheduler()
    entry_future = sched.run(entry())

    sched.tick()
    assert not entry_future.done()
    f.set_result(666)
    assert not entry_future.done()
    sched.tick()
    assert entry_future.done()
    assert entry_future.result() == 666


def test_loop_exc():
    f = Future()

    def get_response():
        return f

    async def entry():
        return await get_response()

    sched = Scheduler()
    entry_future = sched.run(entry())

    sched.tick()
    assert not entry_future.done()
    f.set_exception(KeyError)
    assert not entry_future.done()
    sched.tick()
    assert entry_future.done()
    with pytest.raises(KeyError):
        entry_future.result()


def test_cancel():
    async def entry():
        await suspend()
        assert False

    sched = Scheduler()
    future = sched.run(entry())
    assert future.cancel()
    assert not future.cancelled()
    sched.tick()
    assert future.cancelled()
    assert not future.cancel()
    with pytest.raises(CancelledError):
        future.result()


def test_cancel_self():
    async def entry():
        future.cancel()
        await suspend()
        assert False

    sched = Scheduler()
    future = sched.run(entry())
    sched.tick()
    assert future.cancelled()
    with pytest.raises(CancelledError):
        future.result()


def test_cancel_self_late():
    async def entry():
        await suspend()
        future.cancel()

    sched = Scheduler()
    future = sched.run(entry())
    sched.tick()
    assert future.cancelled()
    with pytest.raises(CancelledError):
        future.result()


def test_cancel_all_tasks():
    async def entry():
        await suspend()
        assert False

    sched = Scheduler()
    sched.run(entry())
    sched.run(entry())
    sched.run(entry())
    sched.cancel_all_tasks()
    sched.tick()


def test_cancel_recursively():
    done_order = []

    fut1 = Future()
    fut2 = Future()

    async def some_recursion(n=10):
        if not n:
            await fut1
        else:
            await some_recursion(n - 1)

    async def entry():
        await some_recursion()
        await fut2

    sched = Scheduler()
    ef = sched.run(entry())

    fut1.add_done_callback(done_order.append)
    ef.add_done_callback(done_order.append)

    sched.tick()
    ef.cancel()
    sched.tick()
    assert fut1.cancelled()
    assert ef.cancelled()
    assert done_order == [fut1, ef]
    assert not fut2.cancelled()


def test_cancel_done_callback():
    done_called_by = None

    def done(fut):
        nonlocal done_called_by
        done_called_by = fut

    async def entry():
        await suspend()

    sched = Scheduler()
    ef = sched.run(entry())
    ef.add_done_callback(done)
    ef.cancel()
    sched.tick()
    assert done_called_by == ef


def test_run_until_all_tasks_finished():
    async def entry():
        await suspend()
        return 1234

    sched = Scheduler()
    future = sched.run(entry())
    sched.run_until_all_tasks_finished()
    assert future.result() == 1234


def test_gather():
    fut1 = Future()
    fut2 = Future()
    fut3 = Future()
    sched = Scheduler()
    gf = gather(fut1, fut2, fut3, scheduler=sched)
    assert not gf.done()
    fut1.set_result(1)
    assert not gf.done()
    fut3.set_result(3)
    fut2.set_result(2)
    assert not gf.done()
    sched.tick()
    assert gf.done()
    assert gf.result() == [1, 3, 2]


def test_gather_exc():
    fut1 = Future()
    fut2 = Future()
    fut3 = Future()
    sched = Scheduler()
    gf = gather(fut1, fut2, fut3, scheduler=sched)
    assert not gf.done()
    fut1.set_result(1)
    assert not gf.done()
    fut3.set_exception(KeyError)
    assert not gf.done()
    sched.tick()
    assert gf.done()
    # fut3 failed, so the other pending futures should enter cancelled state.
    assert fut2.cancelled()
    with pytest.raises(KeyError):
        gf.result()


def test_as_completed():
    fut1 = Future()
    fut2 = Future()
    fut3 = Future()
    sched = Scheduler()
    ac = as_completed(fut1, fut2, fut3, scheduler=sched)

    fut1.set_result(5)
    assert sched.run_until_complete(ac.__anext__()) == 5
    fut2.set_result(1)
    assert sched.run_until_complete(ac.__anext__()) == 1
    fut3.set_result(3)
    assert sched.run_until_complete(ac.__anext__()) == 3


def test_as_completed_exc():
    fut1 = Future()
    fut2 = Future()
    fut3 = Future()
    sched = Scheduler()
    ac = as_completed(fut1, fut2, fut3, scheduler=sched)

    fut1.set_result(5)
    assert sched.run_until_complete(ac.__anext__()) == 5
    fut2.set_exception(KeyError)
    with pytest.raises(KeyError):
        sched.run_until_complete(ac.__anext__())
    assert fut3.cancelled()
    with pytest.raises(StopAsyncIteration):
        sched.run_until_complete(ac.__anext__())



from setuptools import setup, find_packages

description = open('README.rst', 'r', encoding='utf-8').read()

setup(
    name='asynker',

    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    tests_require=[
        'asynker[test]',
    ],
    extras_require={
        'test': [
            'pytest',
        ],
    },

    description='Coroutine scheduler for Python 3.5\'s await syntax',
    long_description=description,

    url='https://github.com/enkore/asynker',
    author='Marian Beermann',
    author_email='asynker@enkore.de',

    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
    ],

    packages=find_packages('src'),
    package_dir={'': 'src'},
)



import collections
import enum
import inspect
import threading
import types


class FutureState(enum.Enum):
    PENDING = 'pending'
    FINISHED = 'finished'
    CANCELLED = 'cancelled'


class CancelledError(Exception):
    """Future/Task was cancelled."""


class Future:
    """
    A future represents the result of some computation that may become available in *drumroll* the future.

    A future starts out in PENDING state.

    Calling set_result(v) transitions it to FINISHED state and
    will resume dependent coroutines (i.e. those waiting on this Future)
    at the next Scheduler.tick().

    Calling set_exception(e) transitions it to FINISHED state as well
    and generally works the same way as set_result(), but instead
    of providing a value to dependent coroutines, they will blow
    up with an exception.

    result() retrieves the current result *or* raises the set exception;
    ``x = await future()`` is semantically equivalent to
    ``twiddle_thumbs_until(future.is.done); x = future.result()``.
    """
    _state: FutureState = FutureState.PENDING
    _result = None
    _exception = None
    _scheduler = None

    def __init__(self, scheduler=None):
        self._scheduler = scheduler
        self._done_callbacks = []

    def done(self):
        """
        A future is done if it (1) finished (with a result or exception) (2) was cancelled.
        """
        return self._state != FutureState.PENDING

    def cancel(self):
        self._exception = CancelledError()
        self._state = FutureState.CANCELLED
        self._schedule_callbacks()
        return True

    def cancelled(self):
        return self._state == FutureState.CANCELLED

    def result(self):
        assert self.done()
        if self._exception is not None:
            raise self._exception
        return self._result

    def exception(self):
        assert self.done()
        return self._exception

    def add_done_callback(self, fn):
        """
        Add callback *fn* that will be run when the future enters a done state.

        The callback receives the futures as its sole argument.
        """
        if self._state != FutureState.PENDING:
            fn(self)
        else:
            self._done_callbacks.append(fn)

    def remove_done_callback(self, fn):
        self._done_callbacks.remove(fn)

    def set_result(self, result):
        assert self._state == FutureState.PENDING
        self._state = FutureState.FINISHED
        self._result = result
        self._schedule_callbacks()

    def set_exception(self, exception):
        assert self._state == FutureState.PENDING
        if isinstance(exception, type):
            exception = exception()
        self._state = FutureState.FINISHED
        self._exception = exception
        self._schedule_callbacks()

    def _schedule_callbacks(self):
        for cb in self._done_callbacks:
            self._scheduler.call_soon(cb, self)

    def __await__(self):
        if not self.done():
            yield self
        return self.result()


class Task(Future):
    """
    A Task is a concretisation  of the Future concept;
    it represents a thread (no, wait, loaded word)... uhm...
    a *line* (fiber? oops...) of execution.

    In other words, a Task object corresponds to a coroutine *invocation*
    (the terminology is pretty loose here, in Python's terms, "async_foo()" *is* a coroutine),
    so it is a concrete computation that can be advanced in response to Futures finishing.

    The result() of a Task is the ``return`` value of the coroutine it represents.
    """

    def __init__(self, coroutine, scheduler):
        super().__init__(scheduler)
        self._coroutine = coroutine
        self._cancel = False
        self._waits_on_future = None

    def cancel(self):
        if self.done():
            return False
        self._cancel = True
        self._scheduler._queue_task(self)
        return True

    def _tick(self, source_future=None):
        if source_future:
            # source_future is "what send us here",
            # i.e. the coroutine awaited some future and
            # *now* that future is completed and called source_future.
            exc = source_future.exception()
            value = source_future._result
        else:
            # When source_future is None, then there are two ways to end up here:
            # (1) the initial run() of a Task was called
            #     (at which point the coroutine is (1) either done because it never yielded
            #      or (2) has reached the first yield point)
            # (2) via Task.cancel() in which case we are here to inject the cancellation
            #     into the coroutine regardless of its progression.
            exc = value = None

        if self._cancel:
            # When we are asked to cancel the execution of the coroutine, we inject
            # a CancelledError exception at the next yield point of the coroutine.
            # This overrides prior exceptions.
            # Handling a CancelledError swallows any result you could've gotten out of an await/yield.
            # So... you probably don't want to do that.

            if self._waits_on_future:
                # If we are supposed to cancel the coroutine but we are waiting on a future,
                # then we pass the cancellation one level lower.
                # This continues recursively (technically we'd process it iteratively) until the
                # lowest-level future which will then be cancelled, produce a CancelledError
                # exception and that exception will bubble the stack up in the opposite direction.
                self._cancel = False
                self._waits_on_future.cancel()
                self._waits_on_future = None
                return
            else:
                exc = CancelledError()
                self._cancel = False

        if self.cancelled():
            return

        self._waits_on_future = None

        try:
            if exc is not None:
                # Inject the exception into the coroutine at the await-point.
                # This can/will trigger the exception stack to wind up,
                # stuff like exception handling and context managers can now happen
                # in the coroutine.
                result = self._coroutine.throw(exc)
            else:
                # Return the value provided by the future to the coroutine,
                # i.e. .send(v) makes the yield-point expression which returned control to us
                # (e.g. an "await" expression somewhere in the call-stack of the coroutine this Task
                # corresponds to) take on the value "v".
                # Note how yielding and resuming can happen at any level in the call-stack of the
                # coroutine; only top-level coroutines have Task objects (via ensure_future() or Scheduler.run()),
                # all other coroutines invoked down the stack are managed by Python on the call stack
                # of the coroutine and not by us (in fact we *can't* take control of them).
                result = self._coroutine.send(value)
        except StopIteration as si:
            # This means the coroutine has returned and is now done.
            if self._cancel:
                # A "late" cancellation, this can pretty much only happen if a coroutine cancels itself
                # after the last yield point in the current control flow path.
                super().cancel()
                self._cancel = False
            else:
                self.set_result(si.value)
        except CancelledError:
            super().cancel()
        except Exception as exc:
            self.set_exception(exc)
        else:
            # "result" can be a couple of things.
            # An "await" works mostly like a "yield from" (and *not* like a "yield" --
            # an await is not necessarily a yield point of a coroutine, it depends on what
            # the awaitable is/does).
            # This means that await will pass values yielded from the awaitable back up
            # the stack, to us (because we resumed the coroutine doing it).
            #
            # "foo = await some_coroutine()" will thus *not* pass a coroutine object to us,
            # it will pass whatever some_coroutine() yield to us. If some_coroutine() never
            # yields, then control never returns to us.
            #
            # "foo = await some_future" (which can also look like calling a coroutine
            # e.g. "await some_regular_function_RETURNING_a_future()") will hand us some_future.
            #
            # If we get a future, we add a done callback to it which will schedule the current coroutine.
            #
            # The third case is we get some arbitrary value. This just means that somewhere
            # a bare yield/await happened (e.g. suspend() is a case of this),
            # which bubbles the value up to us as usual.
            # In that case we immediately re-queue the current coroutine
            # because it can continue at least another iteration.
            if isinstance(result, Future):
                result._scheduler = self._scheduler
                self._wait_on_future(result)
            else:
                self._scheduler._queue_task(self)

    def _wait_on_future(self, future):
        future.add_done_callback(lambda src: self._scheduler._queue_task(self, src))
        self._waits_on_future = future


def ensure_future(future_or_coroutine, scheduler) -> Future:
    """
    Convert *future_or_coroutine* to a Future instance belonging to *scheduler*.

    If *future_or_coroutine* is not a coroutine, it is assumed to be a Future
    and consequently adopted to *scheduler*.
    """
    d = dir(future_or_coroutine)
    # The first case should be obvious
    # The second case is asynchronous generator objects, i.e.
    # async def foo(): yield 1; yield 2; ...
    # => foo() returns an asynchronous generator object
    # The third case is non-obvious but catches various coroutine-like objects,
    # including async_generator_asend-objects, which is, in essence, a single-step coroutine
    # corresponding to one yield of the generator.
    if inspect.iscoroutine(future_or_coroutine) or inspect.isasyncgen(future_or_coroutine) or ('close' in d and 'send' in d and 'throw' in d):
        return Task(future_or_coroutine, scheduler)
    future_or_coroutine._scheduler = scheduler
    return future_or_coroutine


@types.coroutine
def suspend():
    """
    Yield control to the scheduler for one iteration.
    """
    yield


def gather(*futures_or_coroutines, scheduler):
    """
    Wait for all futures and/or coroutines concurrently.

    The await-expression will return a list of results in arbitrary order.
    """
    def done(fut):
        futs.remove(fut)
        if fut.exception():
            gathering_future.set_exception(fut.exception())
            for f in futs:
                f.remove_done_callback(done)
                f.cancel()
        else:
            results.append(fut.result())
        if not futs:
            gathering_future.set_result(results)

    gathering_future = Future()
    futs = set()
    results = []
    for f in futures_or_coroutines:
        f = ensure_future(f, scheduler)
        futs.add(f)
        f.add_done_callback(done)
        scheduler.run(f)
    return gathering_future


async def as_completed(*futures_or_coroutines, scheduler):
    """
    Yields results as they become available.
    """
    def done(fut):
        futs.remove(fut)
        if fut.exception():
            blocker_future.set_exception(fut.exception())
            for f in futs:
                f.remove_done_callback(done)
                f.cancel()
        else:
            results.append(fut.result())
            if not blocker_future.done():
                blocker_future.set_result(None)

    blocker_future = Future()
    futs = set()
    results = []
    for f in futures_or_coroutines:
        f = ensure_future(f, scheduler)
        futs.add(f)
        f.add_done_callback(done)
        scheduler.run(f)

    while futs:
        await blocker_future
        blocker_future = Future()
        res = results
        results = []
        for r in res:
            yield r

    for r in results:
        yield r


class Scheduler:
    def __init__(self):
        self._queue = collections.deque()
        self._tasks = set()

    def tick(self):
        """
        Perform all currently pending actions, e.g. resuming runnable coroutines.
        """
        while self._queue:
            cb, args = self._queue.popleft()
            cb(*args)

    def run_until_complete(self, future_or_coroutine):
        """
        Run *future_or_coroutine* until completion.
        """
        future = ensure_future(future_or_coroutine, self)
        self._queue_task(future)
        while not future.done():
            self.tick()
        return future.result()

    def run_until_all_tasks_finished(self):
        """
        Run until there are no more running tasks/coroutines.
        """
        while self._tasks:
            self.tick()

    def run(self, future_or_coroutine):
        """
        Add *future_or_coroutine* to the running set of futures.
        """
        future = ensure_future(future_or_coroutine, self)
        if isinstance(future, Task):
            self._queue_task(future)
        return future

    def call_soon(self, cb, *args):
        """
        Call *cb* with *args* next time a .tick() happens.
        """
        self._queue.append((cb, args))

    def cancel_all_tasks(self):
        """
        Cancel all currently running tasks (running coroutines).
        """
        for task in self._tasks:
            task.cancel()

    def _queue_task(self, task_future, src=None):
        if task_future not in self._tasks:
            self._tasks.add(task_future)
            task_future.add_done_callback(lambda tf: self._tasks.remove(task_future))
        self._queue.append((task_future._tick, (src,)))



"""
This is a very simplified example showing how Asynker can be used in conjunction
with the Qt event loop.
"""

import sys

from PyQt5.QtCore import QCoreApplication, QObject, QTimer, QEventLoop
from asynker import Scheduler, Future, suspend


app = QCoreApplication(sys.argv)
done = False


def quit():
    global done
    done = True


class Timer(QObject):
    def __init__(self, parent=None):
        super().__init__(parent)

    async def run_event_loop(self):
        timer = QTimer(self)
        timer.start(1000)
        while not done:
            await suspend()
            app.processEvents(QEventLoop.AllEvents | QEventLoop.WaitForMoreEvents)

    def wait(self, ms):
        future = Future()
        timer = QTimer(self)
        timer.timeout.connect(lambda: (future.set_result(None), timer.deleteLater()))
        timer.setSingleShot(True)
        timer.start(ms)
        return future


async def waiterA():
    for n in range(5):
        await timer.wait(500)
        print('waiterA', n)


async def waiterB():
    for n in range(10):
        await timer.wait(100)
        print('waiterB', n)


async def waiterC():
    await timer.wait(3000)
    print('Quitting.')
    quit()


timer = Timer()
scheduler = Scheduler()
scheduler.run(waiterA())
scheduler.run(waiterB())
scheduler.run(waiterC())
scheduler.run_until_complete(timer.run_event_loop())